#include "player_states.hpp"
#include "player_character.hpp"
#include "tunnels_game_mode.hpp"
#include "utils/game_root.hpp"
#include <cmath>
#include <godot_cpp/variant/utility_functions.hpp>

namespace godot {
void PlayerState::_bind_methods() {
#define CLASSNAME PlayerState
}

void PlayerState::init(Node *target, StateMachine *machine) {
    this->state_machine = machine;
    this->parent = Object::cast_to<PlayerCharacter>(target);
    if(!this->parent)
        UtilityFunctions::push_error("Attempt to apply player state to non-PlayerCharacter node of type '", target->get_class(), "'");
    Ref<TunnelsGameMode> game_mode = GameRoot::get_singleton()->get_game_mode();
    this->player = game_mode->get_player_instance();
    if(!this->player)
        UtilityFunctions::push_error("Failed to find valid TunnelsPlayer instance in game mode");
}

StringName PlayerState::get_next() const {
    return this->get_class();
}

PlayerCharacter *PlayerState::get_manual_character() const {
    return this->player->get_character();
}

Vector3 PlayerState::get_home_position() const {
    Basis const player_basis = this->get_manual_character()->get_global_transform().basis;
    Vector3 const player_origin = this->get_manual_character()->get_global_position();
    Vector3 const offset = this->parent->get_global_position() - player_origin;
    Vector3 const local = Vector3(0.f, (std::signbit(player_basis.get_column(0).dot(offset)) ? 1.f : -1.f), -1.f).normalized() * 3.f;
    return player_origin + player_basis.get_column(0) * local.x + player_basis.get_column(2) * local.z;
}

#undef CLASSNAME // PlayerState

void FollowingPlayer::_bind_methods() {
#define CLASSNAME FollowingPlayer
}

void FollowingPlayer::_process(double delta_time) {
    this->parent->move_to(this->get_home_position(), 1.f);
    this->parent->aim_direction(this->parent->get_velocity_target());
}

#undef CLASSNAME // FollowingPlayer
}
